import os
import argparse
import platform

import reCBZ
from reCBZ import wrappers, util
from reCBZ.config import Config


def print_title() -> None:
    align = int(Config.term_width() / 2) - 11
    if align > 21: align = 21
    if align + 22 > Config.term_width() or align < 0:
        align = 0
    align = align * ' '
    title_multiline = (f"{align}┬─┐┌─┐┌─┐┌┐ ┌─┐ ┌─┐┬ ┬\n"
                       f"{align}├┬┘├┤ │  ├┴┐┌─┘ ├─┘└┬┘\n"
                       f"{align}┴└─└─┘└─┘└─┘└─┘o┴   ┴")
    print(title_multiline)


def unix_like_glob(arglist:list) -> list:
    """What we're essentially trying to supplant here is powershell's obtuse
    Get-ChildItem 'pattern' | foreach {cmd $_.FullName}, because it's just
    ridiculous expecting users to memorize that when a simple asterisk could do
    the job. PS if you're using Windows this is a reminder your OS SUCKS ASS"""
    import glob

    new = []
    for arg in arglist:
        if '*' in arg:
            new.extend(glob.glob(str(arg)))
        else:
            new.append(arg)

    return new


def main():
    # o god who art in heaven please guard mine anime girls
    readme='https://github.com/avalonv/reCBZ/blob/master/README.md#usage'
    parser = argparse.ArgumentParser(
            prog=reCBZ.CMDNAME,
            usage="%(prog)s [options] files.cbz",
            epilog=f"for detailed documentation, see {readme}")
    mode_group = parser.add_mutually_exclusive_group()
    fmt_group = parser.add_mutually_exclusive_group()
    ext_group = parser.add_mutually_exclusive_group()
    log_group = parser.add_mutually_exclusive_group()
    process_group = parser.add_mutually_exclusive_group()
    # TODO clean leftover options
    parser.add_argument( "-nw", "--nowrite",
        default=Config.nowrite,
        dest="nowrite",
        action="store_true",
        help="dry run, no changes are saved at the end (safe)")
    ext_group.add_argument( "-O", "--overwrite",
        default=Config.overwrite,
        dest="overwrite",
        action="store_true",
        help="overwrite the original archive")
    parser.add_argument( "-F", "--force",
        default=Config.ignore,
        dest="ignore",
        action="store_true",
        help="ignore file errors when writing pages (dangerous)")
    log_group.add_argument( "-v", "--verbose",
        default=Config.loglevel,
        dest="loglevel",
        action="count",
        help="increase verbosity of progress messages, repeatable: -vvv")
    log_group.add_argument( "-s", "--silent",
        const=-1,
        dest="loglevel",
        action="store_const",
        help="disable all progress messages")
    # mode_group.add_argument( "-u", "--unpack",
    #     const=0,
    #     dest="mode",
    #     action="store_const",
    #     help="unpack the archive to the currect directory (safe)")
    mode_group.add_argument( "-c", "--compare",
        const=1,
        dest="mode",
        action="store_const",
        help="test a small sample with all formats and print the results (safe)")
    mode_group.add_argument( "-a", "--assist",
        const=2,
        dest="mode",
        action="store_const",
        help="compare, then ask which format to use for a real run")
    mode_group.add_argument( "-A" ,"--auto",
        const=3,
        dest="mode",
        action="store_const",
        help="compare, then automatically picks the best format for a real run")
    fmt_group.add_argument( "--nowebp",
        default=Config.blacklistedfmts,
        const=f'{Config.blacklistedfmts} webp webpll',
        dest="blacklistedfmts",
        action="store_const",
        help="exclude webp from --auto and --assist")
    ext_group.add_argument( "--epub",
        default=Config.outformat,
        const='epub',
        dest="outformat",
        action="store_const",
        help="save archive as epub")
    ext_group.add_argument( "--zip",
        default=Config.outformat,
        const='zip',
        dest="outformat",
        action="store_const",
        help="save archive as zip")
    ext_group.add_argument( "--cbz",
        default=Config.outformat,
        const='cbz',
        dest="outformat",
        action="store_const",
        help="save archive as cbz")
    parser.add_argument( "--compress",
        default=Config.compresszip,
        dest="compresszip",
        action="store_true",
        help="attempt to further compress the archive when repacking")
    process_group.add_argument("--process",
        default=Config.processes,
        choices=(range(1,33)),
        metavar="[1-32]",
        dest="processes",
        type=int,
        help="maximum number of processes to spawn")
    process_group.add_argument( "--sequential",
        default=Config.parallel,
        dest="parallel",
        action="store_false",
        help="disable multiprocessing")
    fmt_group.add_argument( "--fmt",
        default=Config.imageformat,
        choices=('jpeg', 'png', 'webp', 'webpll'),
        metavar="fmt",
        dest="imageformat",
        type=str,
        help="format to convert pages to: jpeg, webp, webpll, or png")
    parser.add_argument( "--quality",
        default=Config.quality,
        choices=(range(1,101)),
        metavar="[0-95]",
        dest="quality",
        type=int,
        help="save quality for lossy formats. >90 not recommended")
    parser.add_argument( "--size",
        metavar="WidthxHeight",
        default=Config.resolution,
        dest="resolution",
        type=str,
        help="rescale images to the specified resolution")
    parser.add_argument( "-noup", "--noupscale",
        default=Config.noupscale,
        dest="noupscale",
        action="store_true",
        help="disable upscaling with --size")
    parser.add_argument( "-nodw", "--nodownscale",
        default=Config.nodownscale,
        dest="nodownscale",
        action="store_true",
        help="disable downscaling with --size")
    parser.add_argument( "-bw", "--grayscale",
        default=Config.grayscale,
        dest="grayscale",
        action="store_true",
        help="convert images to grayscale")
    parser.add_argument( "--version",
        dest="show_version",
        action="store_true",
        help="show version and exit")
    parser.add_argument( "--config",
        dest="show_config",
        action="store_true",
        help="show current settings")
    args, unknown_args = parser.parse_known_args()
    # this is probably not the most pythonic way to do this
    # I'm sorry guido-san...
    for key, val in args.__dict__.items():
        if key in Config.__dict__.keys():
            setattr(Config, key, val)
    if args.show_config:
        for key, val in Config.__dict__.items():
            print(f"{key} = {val}")
    if args.show_version:
        print(f'{reCBZ.CMDNAME} v{reCBZ.__version__}')
        exit(0)

    # parse files
    if platform.system() == 'Windows':
        unknown_args = unix_like_glob(unknown_args)
    paths = []
    for arg in unknown_args:
        if os.path.isfile(arg):
            paths.append(arg)
        elif os.path.isdir(arg):
            print(f'{reCBZ.CMDNAME}: {arg}: is a directory')
            parser.print_usage()
            exit(1)
        else:
            parser.print_help()
            print(f'\nunknown file or option: {arg}')
            exit(1)
    if len(paths) <= 0:
        print(f'{reCBZ.CMDNAME}: missing input file (see --help)')
        parser.print_usage()
        exit(1)
    # everything passed
    if reCBZ.SHOWTITLE: print_title()
    mode = args.mode
    for filename in paths:
        try:
            if mode is None:
                wrappers.repack_fp(filename)
            elif mode == 0:
                wrappers.unpack_fp(filename)
            elif mode == 1:
                wrappers.compare_fmts_fp(filename)
            elif mode == 2:
                wrappers.assist_repack_fp(filename)
            elif mode == 3:
                wrappers.auto_repack_fp(filename)
        except InterruptedError:
            continue
        except (KeyboardInterrupt, util.MPrunnerInterrupt):
            print('\nGoooooooooodbye')
            exit(1)


if __name__ == '__main__':
    # catching errors here won't work, presumably because of namespace mangling
    main()



# Copyright 2022 Aleksandar Erkalović, Gwyn Vales
# This program is free software: you can redistribute it and/or modify it under
# the terms of the Affero GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License for more
# details.
# You should have received a copy of the Affero GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
#
# 'ebooklib', authored by Aleksandar Erkalović, is licensed under the AGPL, and we
# are linking to it, therefore AGPL applies to this module as well. You may NOT
# distirubte derivatives using this module (i.e. epub.py) without also
# distributing the source code, even over a network, and any modifications to
# either epub.py or ebooklib must be licensed under the AGPL as well.
# non-commerical private use is explicitly permitted.
# WORK IN PROGRESS
from pathlib import Path
from uuid import uuid4

# *must* use the git version until Aleksandar updates the pypi version
from ebooklib import epub

from reCBZ.util import human_sort, mylog


def single_volume_epub(name:str, pages:list, width='100%', height='100%') -> str:
    # I'm too sleep deprived to figure out why but the order is inverted
    pages = pages[::-1]
    book = epub.EpubBook()

    # attempt to distinguish author / title
    if '-' in name:
        title, author = name.split('-', 1)
    else:
        title = name
        author = 'Made with reCBZ'
    book.set_identifier(str(uuid4()))
    book.set_title(title)
    book.set_language('en')
    book.add_author(author)

    # do its best to sort alphanumerically TODO is it really necessary?
    # pages = human_sort(pages)
    covert_ops = f'cover{pages[0].fmt.ext[0]}'
    book.set_cover(covert_ops, open(pages[0].fp, 'rb').read())

    # one chapter = one page = one image = lotsa bytes
    chapters = []
    for i, page in enumerate(pages, start=1):
        # source_fp = page.fp
        # basename = page.name
        static_dest = f'static/{page.name}'
        mime_type = page.fmt.mime
        size_spec = f'width={page.size[0]} height={page.size[1]}'
        mylog(f'writing {page.fp} to {static_dest} as {mime_type}')

        chapter = epub.EpubHtml(title=f'Page {i}', file_name=f'page_{i}.xhtml',
                                lang='en')
        chapter.content=f'''<html>
                            <head></head>
                            <body>
                              <img src="{static_dest}" {size_spec}'/>
                            </body>
                            </html>'''

        # read bytes
        image_content = open(page.fp, 'rb').read()
        # store read content relative to zip
        static_img = epub.EpubImage(uid=f'image_{i}', file_name=static_dest,
                                    media_type=mime_type, content=image_content)
        book.add_item(chapter)
        book.add_item(static_img)
        chapters.append(chapter)

    # on a per volume basis, use the index of where each volume starts,
    # otherwise just the first page
    book.toc = (chapters[0],)

    # add navigation files
    book.add_item(epub.EpubNcx())
    book.add_item(epub.EpubNav())

    # create spine
    book.spine = ['cover', 'nav', *(page for page in chapters)]

    # write epub file
    source_fp = f'{name}.epub'
    epub.write_epub(source_fp, book, {})
    return source_fp


def multiple_volume_epub(title:str, volumes:list) -> None:
    raise NotImplementedError



import time
from pathlib import Path

import reCBZ
from reCBZ.config import Config
from reCBZ.archive import Archive
from reCBZ.util import human_bytes, pct_change, shorten, mylog


def pprint_fmt_stats(base:tuple, totals:tuple) -> None:
    lines = f'┌─ Disk size ({reCBZ.SAMPLECOUNT}' + \
             ' pages) with present settings:\n'
    # justify to the left and right respectively. effectively the same
    # as using f'{part1: <25} | {part2: >8}\n'
    part1 = f'│   {base[1]}'.ljust(37)
    part2 = f'{human_bytes(base[0])}'.rjust(8)
    lines += f'{part1} {part2} |  0.00%\n'
    for i, total in enumerate(totals):
        if i == len(totals)-1:
            prefix = '└─'
        else:
            prefix = '├─'
        change = pct_change(base[0], total[0])
        part1 = f'{prefix}{i+1} {total[1]}'.ljust(37)
        part2 = f'{human_bytes(total[0])}'.rjust(8)
        lines += f'{part1} {part2} | {change}\n'
    mylog('', progress=True)
    print(lines[0:-1]) # strip last newline


def pprint_repack_stats(source:dict, new:dict, start_t:float) -> None:
    end_t = time.perf_counter()
    elapsed = f'{end_t - start_t:.2f}s'
    max_width = Config.term_width()
    new_size = new['size']
    source_size = source['size']
    name = new['name']
    op = f"{source['type'].upper()} -> {new['type'].upper()}"
    if len(name) > max_width - 36:
        name = shorten(name, width=max_width - 36)
    if new_size > source_size:
        verb = 'INCREASE!'
    else:
        verb = 'decrease'
    change = pct_change(source_size, new_size)
    line1 = f"┌─ {op}: '{name}' completed in {elapsed}"
    line2 = f"└───■■ Source: {human_bytes(source_size)} ■ New: " +\
            f"{human_bytes(new_size)} ■ {change} {verb} ■■"
    mylog('', progress=True)
    print(line1)
    print(line2)


def compare_fmts_fp(fp:str) -> tuple:
    """Run a sample with each image format, return the results"""
    if Config.loglevel >= 0: print(shorten('[i] Analyzing', fp))
    results = Archive(fp).compute_fmt_sizes()
    pprint_fmt_stats(results[0], results[1:])
    return results


def unpack_fp(fp:str) -> None:
    # not implemented yet
    """Unpack the archive, converting all images within
    Returns path to repacked archive"""
    if Config.loglevel >= 0: print(shorten('[i] Unpacking', fp))
    unpacked = Archive(fp).extract()
    for file in unpacked:
        print(file)
    exit(1)


def repack_fp(fp:str) -> str:
    """Repack the archive, converting all images within
    Returns path to repacked archive"""
    if Config.loglevel >= 0: print(shorten('[i] Repacking', fp))
    start_t = time.perf_counter()
    book = Archive(fp)
    book.extract()
    source_pages = len(book.fetch_pages())
    source_stats = {'name':Path(fp).stem,
                    'size':Path(fp).stat().st_size,
                    'type':Path(fp).suffix[1:]}
    book.convert_pages() # page attributes are inherited from Config at init
    new_pages = len(book.fetch_pages())
    discarded = source_pages - new_pages
    if discarded > 0:
        print(f"[!] {discarded} pages couldn't be written")
        if not Config.ignore:
            print('[!] Aborting')
            raise InterruptedError
    if not Config.nowrite:
        if Config.overwrite:
            name = str(Path.joinpath(Path(fp).parents[0], f'{Path(fp).stem}'))
            Path(fp).unlink()
        # elif savedir TODO
        else:
            name = str(Path.joinpath(Path.cwd(), f'{Path(fp).stem} [reCBZ]'))
        results = book.write_archive(Config.outformat, file_name=name)
    else:
        results = fp
    new_stats = {'name':Path(results).stem,
                 'size':Path(results).stat().st_size,
                 'type':Path(results).suffix[1:]}
    pprint_repack_stats(source_stats, new_stats, start_t)
    return results


def assist_repack_fp(fp:str) -> str:
    """Run a sample with each image format, then ask which to repack
    the rest of the archive with
    Returns path to repacked archive"""
    results = compare_fmts_fp(fp)
    options_dic = {i : total[2] for i, total in enumerate(results[1:])}
    metavar = f'[1-{len(options_dic)}]'
    while True:
        try:
            reply = int(input(f"■─■ Proceed with {metavar}: ")) - 1
            selection = options_dic[reply]
            break
        except (ValueError, KeyError):
            print('[!] Ctrl+C to cancel')
            continue
        except KeyboardInterrupt:
            print('[!] Aborting')
            exit(1)
    Config.imageformat = selection
    return repack_fp(fp)


def auto_repack_fp(fp:str) -> str:
    """Run a sample with each image format, then automatically pick
    the smallest format to repack the rest of the archive with
    Returns path to repacked archive"""
    results = Archive(fp).compute_fmt_sizes()
    selection = {"desc":results[1][1], "name":results[1][2]}
    fmt_name = selection['name']
    fmt_desc = selection['desc']
    if Config.loglevel >= 0: print(shorten(f'[i] Proceeding with', fmt_desc))
    Config.imageformat = fmt_name
    return repack_fp(fp)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import time
import tempfile
import shutil
from zipfile import ZipFile, ZIP_DEFLATED, ZIP_STORED, BadZipFile
from multiprocessing.pool import ThreadPool
from functools import partial
from pathlib import Path

from PIL import Image

from reCBZ.formats import *
from reCBZ.config import Config
from reCBZ.util import mylog, MP_run_tasks, SIGNINT_ctrl_c

# TODO:
# include docstrings

class Page():
    def __init__(self, file_name):
        self.fp = Path(file_name)
        self.name = str(self.fp.name)
        self.stem = str(self.fp.stem)
        self._source_fp = self.fp
        self._img:Image.Image
        self._fmt = None
        self._closed = True

    @property
    def fmt(self):
        if self._fmt is not None:
            return self._fmt
        else:
            PIL_fmt = self.img.format
            if PIL_fmt is None:
                raise KeyError(f"Image.format returned None")
            elif PIL_fmt == "PNG":
                return Png
            elif PIL_fmt == "JPEG":
                return Jpeg
            elif PIL_fmt == "WEBP":
                # https://github.com/python-pillow/Pillow/discussions/6716
                with open(self.fp, "rb") as file:
                    if file.read(16)[-1:] == b"L":
                        return WebpLossless
                    else:
                        return WebpLossy
            else:
                raise KeyError(f"'{PIL_fmt}': invalid format")

    @fmt.setter
    def fmt(self, new):
        self._fmt = new

    @property
    def img(self):
        if self._closed:
            self._img = Image.open(self.fp)
            self._closed = False
            return self._img
        else:
            return self._img

    @img.setter
    def img(self, new:Image.Image):
        self._img = new
        self._closed = False

    @property
    def size(self):
        return self.img.size

    def save(self, dest):
        self.fmt.save(self.img, dest)
        self.fp = Path(dest)
        self.name = str(self.fp.name)
        self.stem = str(self.fp.stem)
        self._img.close()
        self._closed = True

    def __reduce__(self):
        # pickle pee. pum pa rum
        # https://stackoverflow.com/q/19855156/
        return (self.__class__, (self.fp, ))


class Archive():
    source_id:str = 'Source'
    new_id:str = ' [reCBZ]'
    temp_prefix:str = f'reCBZCACHE_'
    validbookformats:tuple = ('cbz', 'zip', 'epub', 'mobi')

    def __init__(self, filename:str):
        mylog('Archive: __init__')
        if Path(filename).exists():
            self.source_path:Path = Path(filename)
        else:
            raise ValueError(f"{filename}: invalid path")
        self._source_stem = self.source_path.stem
        self.opt_parallel = Config.parallel
        self.opt_ignore = Config.ignore
        self._zip_compress = Config.compresszip
        self._fmt_blacklist = Config.blacklistedfmts
        self._fmt_samples = Config.samplescount
        self._pages_format = Config.imageformat
        self._pages_quality = Config.quality
        self._pages_size = Config.resolution
        self._pages_bw = Config.grayscale
        self._pages_noup = Config.noupscale
        self._pages_nodown = Config.nodownscale
        self._pages_filter = Config.resamplemethod
        self._index:list = []
        self.tempdir:Path = Path('.')

    def fetch_pages(self):
        if len(self._index) == 0:
            self._index = list(self.extract())
        return self._index

    def extract(self, count:int=0, raw:bool=False) -> tuple:
        # check and clean previous tempdirs
        prev_dirs = Path(tempfile.gettempdir()).glob(f'{self.temp_prefix}*')
        for path in prev_dirs:
            assert path != tempfile.gettempdir() # for the love of god
            mylog(f'{path} exists, cleaning up')
            shutil.rmtree(path)

        self.tempdir = Path(tempfile.mkdtemp(prefix=f'{self.temp_prefix}'))
        try:
            source_zip = ZipFile(self.source_path)
        except BadZipFile as err:
            raise ValueError(f"Fatal: '{self.source_path}': not a zip file")

        compressed_files = source_zip.namelist()
        assert len(compressed_files) >= 1, 'no files in archive'
        if count > 0:
            # select x images from the middle of the archive, in increments of 2
            if count * 2 > len(compressed_files):
                raise ValueError(f"{self.source_path} is smaller than samples * 2")
            delta = int(len(compressed_files) / 2)
            compressed_files = compressed_files[delta-count:delta+count:2]

        mylog(f'Extracting: {self.source_path}', progress=True)
        for file in compressed_files:
            source_zip.extract(file, self.tempdir)

        # god bless you Georgy https://stackoverflow.com/a/50927977/
        raw_files = tuple(filter(Path.is_file, Path(self.tempdir).rglob('*')))
        pages = tuple(Page(path) for path in raw_files)

        mylog('', progress=True)
        if raw: return raw_files
        else: return pages

    def write_archive(self, book_format='cbz', file_name:str='') -> str:
        if book_format not in Archive.validbookformats:
            raise ValueError(f"Invalid format '{book_format}'")
        if file_name != '':
            parent = Path(file_name).parents[0]
            if not (parent.exists() and parent.is_dir()):
                raise ValueError(f"Parent folder '{parent}' does not exist")
            new_path = Path(f'{file_name}.{book_format}')
        else:
            # write to current dir
            new_path = Path(f'{self._source_stem}.{book_format}')
        if new_path.exists():
            mylog(f'Write .{book_format}: {new_path}', progress=True)
            mylog(f'{new_path} exists, removing...')
            new_path.unlink()

        new_path = str(new_path)
        if book_format == 'cbz':
            return self._write_zip(new_path)
        elif book_format == 'zip':
            return self._write_zip(new_path)
        elif book_format == 'epub':
            return self._write_epub(new_path)
        elif book_format == 'mobi':
            raise NotImplementedError
        else:
            raise ValueError

    def convert_pages(self, fmt=None, quality=None, grayscale=None, size=None) -> tuple:
        if fmt is not None: self._pages_format = fmt
        if quality is not None: self._pages_quality = int(quality)
        if grayscale is not None: self._pages_bw = bool(grayscale)
        if size is not None: self._pages_size = size

        source_pages = self.fetch_pages()
        if self.opt_parallel:
            results = MP_run_tasks(self._convert_page, source_pages)
        else:
            results = map(self._convert_page, source_pages)
        self._index = [page for page in results if page]
        return tuple(self._index)

    def compute_fmt_sizes(self) -> tuple:
        def compute_single_fmt(sample_pages, tempdir, fmt) -> tuple:
            fmtdir = Path.joinpath(tempdir, fmt.name)
            Path.mkdir(fmtdir)

            pfunc = partial(self._convert_page, savedir=fmtdir, format=fmt)
            if self.opt_parallel:
                results = MP_run_tasks(pfunc, sample_pages)
            else:
                results = map(pfunc, sample_pages)

            converted_pages = [page for page in results if page]
            nbytes = sum(page.fp.stat().st_size for page in converted_pages)
            return nbytes, fmt.desc, fmt.name

        # extract images and compute their original size
        # manually call extract so we don't overwrite _pages cache
        source_pages = self.extract(count=self._fmt_samples)
        nbytes = sum(page.fp.stat().st_size for page in source_pages)
        source_fmt = source_pages[0].fmt
        source_fsize = [nbytes, f'{Archive.source_id} ({source_fmt.desc})',
                        source_fmt.name]

        # compute the size of each format after converting.
        # one thread per individual format. n processes per thread
        fmt_fsizes = []
        pfunc = partial(compute_single_fmt, source_pages, self.tempdir)
        if self.opt_parallel:
            with ThreadPool(processes=len(self._valid_page_formats)) as Tpool:
                fmt_fsizes.extend(Tpool.map(pfunc, self._valid_page_formats))
        else:
            fmt_fsizes.extend(map(pfunc, self._valid_page_formats))

        # finally, compare
        # in multidepth lists, sorted compares the first element by default :)
        sorted_fmts = list(sorted(fmt_fsizes))
        sorted_fmts.insert(0, source_fsize)
        mylog(str(sorted_fmts))
        mylog('', progress=True)
        return tuple(sorted_fmts)

    def _write_zip(self, savepath):
        new_zip = ZipFile(savepath,'w')
        for page in self.fetch_pages():
            try:
                dest = page.fp.relative_to(self.tempdir)
                if self._zip_compress:
                    new_zip.write(page.fp, dest, ZIP_DEFLATED, 9)
                else:
                    new_zip.write(page.fp, dest, ZIP_STORED)
            except ValueError:
                msg = 'Path is being screwy. Does tempdir exist? '
                msg += str(self.tempdir.exists())
                msg += '\nwe might not have joined paths correctly in trans img'
                raise ValueError(msg)
        new_zip.close()
        return savepath

    def _write_epub(self, savepath):
        from reCBZ.epub import single_volume_epub
        title = self._source_stem
        mylog(f'Write .epub: {title}.epub', progress=True)
        savepath = single_volume_epub(title, self.fetch_pages())
        return savepath

    def _write_mobi(self, savepath):
        # TODO not implemented
        pass

    def _create_pages_from_path(self, fp):
        # TODO make public as add_page, append to self._pages
        return Page(fp)

    @SIGNINT_ctrl_c
    def _convert_page(self, source:Page, savedir=None, format=None): #-> None | Str:
        start_t = time.perf_counter()
        LossyFmt.quality = self._pages_quality
        # page = copy.deepcopy(source)
        page = Page(source.fp) # create a copy

        try:
            # ensure file can be opened as image
            mylog(f'Read img: {page.name}', progress=True)
            log_buff = f'/open:  {page.fp}\n'
            source_fmt = page.fmt
            img = page.img
        except IOError as err:
            if self.opt_ignore:
                mylog(f"{page.fp}: can't open file as image, ignoring...'")
                return None
            else:
                raise err
        except KeyError as err:
            if self.opt_ignore:
                mylog(f"{page.fp}: invalid image format, ignoring...'")
                return None
            else:
                raise err

        # determine target format
        if format:
            new_fmt = format
        elif self._new_page_format is not None:
            new_fmt = self._new_page_format
        else:
            new_fmt = source_fmt
        page.fmt = new_fmt

        # apply format specific actions
        if new_fmt is Jpeg:
          if not img.mode == 'RGB':
              log_buff += '|trans: mode RGB\n'
              img = img.convert('RGB')

        # transform
        if self._pages_bw:
            log_buff += '|trans: mode L\n' # me lol
            img = img.convert('L')
        if all(self._new_page_size):
            log_buff += f'|trans: resize to {self._new_page_size}\n'
            width, height = img.size
            new_size = self._new_page_size
            # preserve aspect ratio for landscape images
            if width > height:
                new_size = new_size[::-1]
            n_width, n_height = new_size
            # downscaling
            if (width > n_width and height > n_height
                and not self._pages_nodown):
                img = img.resize((new_size), self._pages_filter)
            # upscaling
            elif not self._pages_noup:
                img = img.resize((new_size), self._pages_filter)

        # save
        page.img = img
        ext = page.fmt.ext[0]
        if savedir:
            new_fp = Path.joinpath(savedir, f'{page.stem}{ext}')
        else:
            new_fp = Path.joinpath(page.fp.parents[0], f'{page.stem}{ext}')
        log_buff += f'|trans: {source_fmt.name} -> {new_fmt.name}\n'
        page.save(new_fp)

        end_t = time.perf_counter()
        elapsed = f'{end_t-start_t:.2f}s'
        mylog(f'{log_buff}\\write: {new_fp}: took {elapsed}')
        mylog(f'Save img: {new_fp.name}', progress=True)
        return page

    @property
    def _new_page_format(self):
        if self._pages_format in (None, ''): return None
        elif self._pages_format == 'jpeg': return Jpeg
        elif self._pages_format == 'png': return Png
        elif self._pages_format == 'webp': return WebpLossy
        elif self._pages_format == 'webpll': return WebpLossless
        else: raise ValueError(f"Invalid format name '{self._pages_format}'")

    @property
    def _valid_page_formats(self) -> tuple:
        all_fmts = (Png, Jpeg, WebpLossy, WebpLossless)
        try:
            blacklist = self._fmt_blacklist.lower().split(' ')
        except AttributeError: # blacklist is None
            return all_fmts
        valid_fmts = tuple(fmt for fmt in all_fmts if fmt.name not in blacklist)
        assert len(valid_fmts) >= 1, "valid_formats is 0"
        return valid_fmts

    @property
    def _new_page_size(self) -> tuple:
        default_value = (0,0)
        newsize = self._pages_size.lower().strip()
        try:
            newsize = tuple(map(int,newsize.split('x')))
            assert len(newsize) == 2
            return newsize
        except (ValueError, AssertionError):
            return default_value



from importlib import resources
try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

__version__ = "0.5.0"
CMDNAME = 'recbz'

_cfg = tomllib.loads(resources.read_text("reCBZ", "defaults.toml"))
OVERWRITE = _cfg["archive"]["overwrite"]
IGNORE = _cfg["archive"]["ignore"]
LOGLEVEL = _cfg["archive"]["loglevel"]
PARALLEL = _cfg["archive"]["parallel"]
PROCESSES = _cfg["archive"]["processes"]
OUTFORMAT = _cfg["archive"]["outformat"]
COMPRESSZIP = _cfg["archive"]["compresszip"]
SAMPLECOUNT = _cfg["archive"]["samplecount"]
NOWRITE = _cfg["archive"]["nowrite"]
BLACKLISTEDFMTS = _cfg["archive"]["blacklistedfmts"]
IMAGEFORMAT = _cfg["archive"]["imageformat"]
QUALITY = _cfg["archive"]["quality"]
RESOLUTION = _cfg["archive"]["resolution"]
NOUPSCALE = _cfg["archive"]["noupscale"]
NODOWNSCALE = _cfg["archive"]["nodownscale"]
GRAYSCALE = _cfg["archive"]["grayscale"]
SHOWTITLE = _cfg["archive"]["showtitle"]



import os

from PIL import Image

import reCBZ


class Config():
    overwrite:bool = reCBZ.OVERWRITE
    ignore:bool = reCBZ.IGNORE
    nowrite:bool = reCBZ.NOWRITE
    loglevel:int = reCBZ.LOGLEVEL
    parallel:bool = reCBZ.PARALLEL
    processes:int = reCBZ.PROCESSES
    outformat:str = reCBZ.OUTFORMAT
    compresszip:int = reCBZ.COMPRESSZIP
    samplescount:int = reCBZ.SAMPLECOUNT
    blacklistedfmts:str = reCBZ.BLACKLISTEDFMTS
    imageformat:str = reCBZ.IMAGEFORMAT
    quality:int = reCBZ.QUALITY
    resolution:str = reCBZ.RESOLUTION
    noupscale:bool = reCBZ.NOUPSCALE
    nodownscale:bool = reCBZ.NODOWNSCALE
    grayscale:bool = reCBZ.GRAYSCALE
    # LANCZOS sacrifices performance for optimal upscale quality
    resamplemethod = Image.Resampling.LANCZOS

    @classmethod
    def pcount(cls) -> int:
        default_value = 4
        if cls.processes > 0:
            return cls.processes
        else:
            logical_cores = os.cpu_count()
            try:
                assert logical_cores is not None
                if logical_cores > 1:
                    logical_cores -= 1 # be kind
                return logical_cores
            except AssertionError:
                return default_value


    @classmethod
    def term_width(cls) -> int:
        # limit output message width. ignored if verbose
        TERM_COLUMNS, TERM_LINES = os.get_terminal_size()
        try:
            assert TERM_COLUMNS > 0 and TERM_LINES > 0
            if TERM_COLUMNS > 120: max_width = 120
            elif TERM_COLUMNS < 30: max_width = 30
            else: max_width = TERM_COLUMNS - 2
        except AssertionError:
            print("[!] Can't determine terminal size, defaulting to 78 cols")
            max_width = 78
        return max_width
